@@ -79,8 +79,8 @@ static void init_modules (void);
 /**
  * pango_find_map:
  * @language: the language tag for which to find the map
- * @engine_type_id: the render type for the map to find
- * @render_type_id: the engine type for the map to find
+ * @engine_type_id: the engine type for the map to find
+ * @render_type_id: the render type for the map to find
  * 
  * Locate a #PangoMap for a particular engine type and render
  * type. The resulting map can be used to determine the engine
@@ -163,7 +163,7 @@ pango_engine_pair_get_engine (PangoEnginePair *pair)
 	      return NULL;
 	    }
 	  
-	  g_module_symbol (module, "script_engine_load", (gpointer)&load);
+	  g_module_symbol (module, "script_engine_load", (gpointer *) &load);
 	  if (!load)
 	    {
 	      fprintf(stderr, "cannot retrieve script_engine_load from %s: %s\n",
@@ -244,6 +244,28 @@ process_module_file (FILE *module_file)
 	    {
 	    case 0:
 	      pair->load_info = g_strdup (tmp_buf->str);
+#if defined(G_OS_WIN32) && defined(LIBDIR)
+	      if (strncmp (pair->load_info,
+			   LIBDIR "/pango/modules/",
+			   strlen (LIBDIR "/pango/modules/")) == 0)
+		{
+		  /* This is an entry put there by make install on the
+		   * packager's system. On Windows a prebuilt Pango
+		   * package can be installed in a random
+		   * location. The pango.modules file distributed in
+		   * such a package contains paths from the package
+		   * builder's machine. Replace the path with the real
+		   * one on this machine. */
+		  gchar *tem = pair->load_info;
+		  pair->load_info =
+		    g_strconcat (pango_get_lib_subdirectory (),
+				 "\\modules\\",
+				 tem + strlen (LIBDIR "/pango/modules/"),
+				 NULL);
+		  g_free (tem);
+		}
+
+#endif
 	      break;
 	    case 1:
 	      pair->info.id = g_strdup (tmp_buf->str);
@@ -498,9 +520,9 @@ build_map (PangoMapInfo *info)
  * @wc:  an ISO-10646 codepoint
  * 
  * Returns the entry in the map for a given codepoint. The entry
- * contains information about engine that should be used for
+ * contains information about the engine that should be used for
  * the codepoint and also whether the engine matches the language
- * tag for the map was created exactly or just approximately.
+ * tag for which the map was created exactly or just approximately.
  * 
  * Return value: the #PangoMapEntry for the codepoint. This value
  *   is owned by the #PangoMap and should not be freed.
