@@ -22,9 +22,6 @@
 #include "config.h"
 
 #include <glib.h>
-#ifdef HAVE_DIRENT_H
-#include <dirent.h>
-#endif
 #include <gmodule.h>
 #include "pango-break.h"
 #include "pango-context.h"
@@ -40,13 +37,6 @@
 
 #ifdef G_OS_WIN32
 #define SOEXT ".dll"
-#ifndef PATH_MAX
-#include <stdlib.h>
-#define PATH_MAX _MAX_PATH
-#endif /* PATH_MAX */
-#include <direct.h>		/* for getcwd() with MSVC */
-#include <io.h>			/* for getcwd() with mingw */
-#define getcwd _getcwd
 #else
 #define SOEXT ".so"
 #endif
@@ -61,10 +51,10 @@ query_module (const char *dir, const char *name)
   GModule *module;
   gchar *path;
 
-  if (name[0] == G_DIR_SEPARATOR)
+  if (g_path_is_absolute (name))
     path = g_strdup (name);
   else
-    path = g_strconcat (dir, G_DIR_SEPARATOR_S, name, NULL);
+    path = g_build_filename (dir, name, NULL);
   
   module = g_module_open (path, 0);
 
@@ -72,9 +62,9 @@ query_module (const char *dir, const char *name)
     fprintf(stderr, "Cannot load module %s: %s\n", path, g_module_error());
 	  
   if (module &&
-      g_module_symbol (module, "script_engine_list", (gpointer)&list) &&
-      g_module_symbol (module, "script_engine_load", (gpointer)&load) &&
-      g_module_symbol (module, "script_engine_unload", (gpointer)&unload))
+      g_module_symbol (module, "script_engine_list", (gpointer *) &list) &&
+      g_module_symbol (module, "script_engine_load", (gpointer *) &load) &&
+      g_module_symbol (module, "script_engine_unload", (gpointer *) &unload))
     {
       gint i,j;
       PangoEngineInfo *engines;
@@ -126,7 +116,7 @@ query_module (const char *dir, const char *name)
 
 int main (int argc, char **argv)
 {
-  char cwd[PATH_MAX];
+  char *cwd;
   int i;
   char *path;
 
@@ -141,9 +131,9 @@ int main (int argc, char **argv)
       
       path = pango_config_key_get ("Pango/ModulesPath");
       if (!path)
-	path = g_strconcat (pango_get_lib_subdirectory (),
-			    G_DIR_SEPARATOR_S "modules",
-			    NULL);
+	path = g_build_filename (pango_get_lib_subdirectory (),
+				 "modules",
+				 NULL);
 
       printf ("# ModulesPath = %s\n#\n", path);
 
@@ -151,28 +141,30 @@ int main (int argc, char **argv)
 
       for (i=0; dirs[i]; i++)
 	{
-	  DIR *dir = opendir (dirs[i]);
+	  GDir *dir = g_dir_open (dirs[i], 0, NULL);
 	  if (dir)
 	    {
-	      struct dirent *dent;
+	      const char *dent;
 
-	      while ((dent = readdir (dir)))
+	      while ((dent = g_dir_read_name (dir)))
 		{
-		  int len = strlen (dent->d_name);
-		  if (len > 3 && strcmp (dent->d_name + len - strlen (SOEXT), SOEXT) == 0)
-		    query_module (dirs[i], dent->d_name);
+		  int len = strlen (dent);
+		  if (len > 3 && strcmp (dent + len - strlen (SOEXT), SOEXT) == 0)
+		    query_module (dirs[i], dent);
 		}
 	      
-	      closedir (dir);
+	      g_dir_close (dir);
 	    }
 	}
     }
   else
     {
-      getcwd (cwd, PATH_MAX);
+      cwd = g_get_current_dir ();
       
       for (i=1; i<argc; i++)
 	query_module (cwd, argv[i]);
+
+      g_free (cwd);
     }
   
   return 0;
